/*********** mtrr_drv.c *****************/ 

#include <KernelExport.h> 
#include <Drivers.h> 
#include <Errors.h> 

#include <OS.h> 
#include <perfmon_kernel.h> 

#include "mtrr_drv.h" 


int    n_cpus; 
sem_id  mtrr_drv_lock_sem; 
sem_id  main_sem;  /* acquired by the main thread, 
                      released by the work threads */ 
sem_id  work_sem;  /* acquired by the work threads, 
                      released by the main thread */ 
volatile long  work_begin_spinlock = 0, 
               work_end_spinlock = 0; 
/* for work threads synchronization with disabled interrupts */ 
volatile long  n_errors;  /* to count work threads errors */ 

mtrr_drv_command_t  kernel_mtrr_command; 

int32 work_thread(void* mtrr_command); 

static status_t 
  set_memory_type(mtrr_drv_command_t* mtrr_command); 
static status_t 
  set_memory_range_type(const mtrr_drv_command_t* mtrr_command); 

static status_t 
  set_mtrr(int64 base, uint64 len, MEM_TYPE mem_type, 
           uint64 mtrr0, int num_mtrrs); 
static status_t 
  clear_mtrr(int64 base, uint64 mtrr0, int num_mtrrs); 

static status_t 
lock_mtrr_drv(void) 
{ 
  return acquire_sem(mtrr_drv_lock_sem); 
} 

static status_t 
unlock_mtrr_drv(void) 
{ 
  return release_sem(mtrr_drv_lock_sem); 
} 


/* 
Only Pentium Pro and Pentium II have memory range type 
registers. Tell the kernel to load the driver only on a 
system with the right CPU. The driver is very conservative 
and does not do anything remotely dangerous, like read CPU 
model-specific registers in init_hardware(). A user has to 
explicitly load/open/read/ioctl this driver to touch the 
hardware. 
*/ 

status_t 
init_hardware (void) 
{ 
//unfortunately I only know the ID of _one_ VIA cpu :-/
/*
  system_info  sys_info; 

  get_system_info(&sys_info); 
  switch (sys_info.cpu_type) 
  { 
  case 0x1369:// VIA C3 Nehemiah 
    return B_OK; 
  default: 
    return B_ERROR; 
  } 
*/

	//assuming all VIA CPU's have same stuff, otherwise we need to diff for 
	//family/stepping... (advice from the VIA CPU spec docs)
	cpuid_info id_inf;
	uint32 eax = 0x00;
	uint8 cnt;
	char vendorid[12] = "CentaurHauls";

	get_cpuid(&id_inf, eax, 0);
	for (cnt = 0 ; cnt < 12; cnt ++)
	{
		if (id_inf.eax_0.vendorid[cnt] != vendorid[cnt]) return B_ERROR;
	}
	return B_OK;
} 


status_t 
init_driver (void) 
{ 
  if((mtrr_drv_lock_sem = 
      create_sem(1, "mtrr_drv_lock")) < B_OK) 
    return B_ERROR; 

  return B_OK; 
} 

void 
uninit_driver (void) 
{ 
  delete_sem(mtrr_drv_lock_sem); 
} 

static status_t 
mtrr_drv_open (const char *name, uint32 flags, void** cookie) 
{ 
  return B_OK; 
} 

static status_t 
mtrr_drv_read (void* cookie, off_t position, void *buf, 
               size_t* num_bytes) 
{ 
  *num_bytes = 0;     /* tell caller nothing was read */ 
  return B_IO_ERROR; 
} 

static status_t 
mtrr_drv_write (void* cookie, off_t position, 
                const void* buffer, size_t* num_bytes) 
{ 
  *num_bytes = 0;     /* tell caller nothing was written */ 
  return B_IO_ERROR; 
} 

static status_t 
mtrr_drv_control (void* cookie, uint32 ioctl, void* arg, 
                  size_t len) 
{ 
  status_t  res; 

  lock_mtrr_drv(); 
  switch(ioctl) 
  { 
  case MTTR_DRV_COMMAND: 
    res = set_memory_type((mtrr_drv_command_t*) arg); 
//res = B_OK;
    break; 

  default: 
    res= B_BAD_VALUE; 
  } 
  unlock_mtrr_drv(); 
  return res; 
} 

static status_t 
mtrr_drv_close (void* cookie) 
{ 
  return B_OK; 
} 

static status_t 
mtrr_drv_free (void* cookie) 
{ 
  return B_OK; 
} 

static const char *mtrr_drv_name[] = { 
  "cpu/mtrr", 
  NULL 
}; 

device_hooks mtrr_drv_hooks = { 
  mtrr_drv_open, 
  mtrr_drv_close, 
  mtrr_drv_free, 
  mtrr_drv_control, 
  mtrr_drv_read, 
  mtrr_drv_write 
}; 

const char** 
publish_devices() 
{ 
  return mtrr_drv_name; 
} 

device_hooks* 
find_device(const char* name) 
{ 
  return &mtrr_drv_hooks; 
} 

/* Access to x86 control registers */ 

static void 
set_cr0(uint32 val) 
{ 
  asm ("mov %%cr0, %%eax"
  :
  : "a" (val)
  ); 
} 

static uint32 
get_cr0(void) 
{ 
  uint32  res = 0; 

  asm ("mov %%eax, %%cr0"
  		: "=a" (res) 	//outputs
  		:				//inputs
  		); 		//used registers
  return res; 
} 

static void 
set_cr3(uint32 val) 
{ 
  asm ("mov %%cr3, %%eax"
    	:
    	: "a" (val)
    	); 
} 

static uint32 
get_cr3(void) 
{ 
  uint32  res = 0; 

  asm ("mov %%eax, %%cr3"
  		: "=a" (res)
  		:
  		); 
  return res; 
} 

static uint32 
get_cpu_features(void) 
{ 
  uint32  res = 0; 

  asm (" 
    mov $0x00000001, %%eax 
    cpuid" 
    : "=d" (res)//outputs (works..)
    :
	); 

  return res; 
} 

/* roundup the arg to the nearest 2 to the nth power */ 
static uint64 
roundup(uint64 val) 
{ 
  int i; 
  uint64  roundup_val; 
  uint64  tmp_val; 

  for(i=0, tmp_val = val - 1; tmp_val != 0; tmp_val >>= 1, i++) 

    ; 
  roundup_val = 1 << i; 

  return roundup_val; 
} 

static bool 
is_aligned(uint64 addr, uint64 alignment) 
{ 
  do 
  { 
    alignment >>= 1; 
    if(addr & alignment) 
      return FALSE; 
  } 
  while(alignment != 0); 

  return TRUE; 
} 

static status_t 
set_memory_type(mtrr_drv_command_t* mtrr_command) 
{ 
  system_info      sys_info; 
  physical_entry   phys_mem_start; 
  int              i; 

  /* dprintf("%d %d %Lx %Lx\n", mtrr_command->action, 
     mtrr_command->mem_type, 
     mtrr_command->base, mtrr_command->len); */ 

  /* do arguments checking and conditioning  */ 
  /* allow only supported memory types       */ 
  if(mtrr_command->action == RANGE_SET) 
	  switch(mtrr_command->mem_type) 
	  { 
    	case UC_MEM :  /* uncachable       */ 
  		case WC_MEM :  /* write combining  */ 
    	case WT_MEM :  /* write-through    */ 
    	case WP_MEM :  /* write-protected  */ 
    	case WB_MEM :  /* writeback        */ 
    	  break; 

    	default: 
      		return B_BAD_VALUE; 
  		} 

  /* 
  len has to be a power of 2 and base has to be aligned 
  on at least len boundary. The minimum len is 4 KByte. 
  */ 
  if(mtrr_command->action == RANGE_SET) 
  { 
    mtrr_command->len = roundup(mtrr_command->len); 
    if(mtrr_command->len < 4096) 
      return B_BAD_VALUE; 
  } 

  /* 
   The driver gets virtual base address but mtrrs need 
   physical address. Convert virtual to physical. In error 
   case assume that physical address was passed. 
  */ 
  get_memory_map(((const void*)&(mtrr_command->base)), 4096, 
                 &phys_mem_start, 1); 

  if(phys_mem_start.size >= 4096) 
    /* this virtual address is mapped */ 
    mtrr_command->base = (uint64)(((uint32)((uint32*)(phys_mem_start.address)))); 

  if(mtrr_command->action == RANGE_SET) 
  { 
    if(!is_aligned(mtrr_command->base, mtrr_command->len)) 
      return B_BAD_VALUE; 
  } 

  get_system_info(&sys_info); 
  n_cpus = sys_info.cpu_count; 

  /* create a semaphore for the main thread to wait for 
     all work threads to start/finish */ 
  main_sem = create_sem(0, "mtrr_drv_main"); 

  /* create a semaphore for work threads to wait to start 
     simultaneously */ 
  work_sem = create_sem(0, "mtrr_drv_work"); 

  /* 
   copy data from the user space to the kernel space: kernel 
   threads do not have access to the user space. 
  */ 
  memcpy(&kernel_mtrr_command, mtrr_command, 
         sizeof(kernel_mtrr_command)); 

  /* spawn a kernel thread for each cpu in the system */ 
  for(i=0; i< n_cpus; i++) 
  { 
    char thread_name[] = "mtrr_x"; 

    thread_name[5]=i+'0'; 
    resume_thread(spawn_kernel_thread(work_thread, 
      thread_name, B_REAL_TIME_PRIORITY, 
      (void*)&kernel_mtrr_command)); 
  } 

  /* reset the work threads error counter */ 
  n_errors = 0; 

  /* wait for all work threads to start */ 
  acquire_sem_etc(main_sem, n_cpus, 0, 0); 

  /* release all work threads together. Big Bang!*/ 
  release_sem_etc(work_sem, n_cpus, 0); 
  /* All work threads are running on all CPUs in the system */ 

  /* wait until all work threads have done the job */ 
  acquire_sem_etc(main_sem, n_cpus, 0, 0); 

  /* give some time to work threads to die */ 
  snooze(10*1000); 

  dprintf("n_errors = %d\n", n_errors); 
  /* Done! */ 

  /* do not leak semaphores */ 
  delete_sem(work_sem); 
  delete_sem(main_sem); 

  return (n_errors == 0) ? B_OK : B_ERROR; 
} 


int32 
work_thread(void* arg) 
{ 
  cpu_status  ps; 
  int32       res; 
  const mtrr_drv_command_t* mtrr_command = 
    (const mtrr_drv_command_t*) arg; 

  /* tell the main thread that this thread started */ 
  release_sem(main_sem); 

  /* wait the beginning of the Big Bang */ 
  acquire_sem(work_sem); 

  /* Each work thread is running on its own CPU. Disable 
     scheduling and thus lock work threads on CPUs */ 
  ps = disable_interrupts(); 
  /* 
   Yes! This thread owns the current CPU until 
   restore_interrupts() or any scheduling calls 
   (acquire/release_sem(), etc.). Start multiple-processor 
   MTTR change protocol. See Intel Architecture Software 
   Developer's Manual, Volume 3: System Programming Guide, 
   chapter 9.11.8 for details. 
   http://www.intel.com/design/pentiumii/manuals/243192.htm 
  */ 

  /* 
  Busy wait for other CPUs/work threads to reach this point. 
  A thread can not call acquire_sem() or any potentially blocking calls 
  if interrupts are disabled. 
  */ 

  /* tell other work threads: "I am here." */ 
  atomic_add(&work_begin_spinlock, 1); 

  /* Busy wait. work_spinlock is volatile so the compiler 
     will not optimize out this loop */ 
  while(work_begin_spinlock != n_cpus) 
    ; 

  /* Do the real work */ 
  res = set_memory_range_type(mtrr_command); 
//res = B_OK;

  /* tell other work threads: "I am here." */ 
  atomic_add(&work_end_spinlock, 1); 
  /* Busy wait. */ 
  while(work_end_spinlock != n_cpus) 
    ; 

  /* return to the normal mode */ 
  restore_interrupts(ps); 

  /* update error counter */ 
  if(res != B_OK) 
    atomic_add(&n_errors, 1); 

  /* Tell the main tread: "I am done!" */ 
  release_sem(main_sem); 

  return res; 
} 



/* ---------- 
  set_mtrrs - scans the variable MTRRs to find already 
  programmed MTRR. If such MTTRs is not found, find the 
  first unused and program it to use mem_type for memory 
  with starting address base and length len. len has to be 
  a power of 2 and base has to be aligned on at least len 
  boundary. The minimum len is 4 KByte. 

  Returns B_OK if the MTRR is set or already has been set. 
  Returns B_ERROR if all MTRRs are in use 
----- */ 

static status_t 
set_mtrr(int64 base, uint64 len, MEM_TYPE mem_type, 
         uint64 mtrr0, int num_mtrrs) 
{ 
  uint64 mask; 
  uint64 mtrr; 

  mask = len - 1; 
  mask = ~mask; 
  mask &= 0x0000000FFFFFF000ULL; 
  mask |= (1 << 11); /* set enable bit */ 

  base = (base & 0xFFFFFF000ULL) | mem_type; 

  for(mtrr = mtrr0; mtrr < mtrr0 + num_mtrrs*2; mtrr += 2) 
  { 
    if(read_msr(mtrr) == base) 
    { 
      if(read_msr(mtrr + 1) == mask) 
        /* BIOS already has set this memory range 
           with our type */ 
      { 
        return B_OK; 
      } 
      else  /* BIOS setup is broken, reprogram this register */ 
      { 
        write_msr(mtrr,  base); 
        write_msr(mtrr+1, mask); 
        return B_OK; 
      } 
    } 
  } 

  /* find an unused MTRR */ 
  for(mtrr = mtrr0; mtrr < mtrr0 + num_mtrrs*2; mtrr += 2) 
  { 
    if(!(read_msr(mtrr+1) & (1 << 11) )) 
      /* the register is disabled, we can use it! */ 
      break; 
  } 
  if(mtrr == (mtrr0 + num_mtrrs*2)) 
    /* all regs are in use, we can't do anything */ 
    return B_ERROR; 

  write_msr(mtrr,  base); 
  write_msr(mtrr+1, mask); 
  return B_OK; 
} 

static status_t 
clear_mtrr(int64 base, uint64 mtrr0, int num_mtrrs) 
{ 
  uint64 mtrr; 

  base = base & 0xFFFFFF000ULL; 

  for(mtrr = mtrr0; mtrr < mtrr0 + num_mtrrs*2; mtrr += 2) 
  { 
    if((read_msr(mtrr) & 0xFFFFFF000ULL ) == base) 
    { 
      write_msr(mtrr+1, 0); /* clear enable bir */ 
      write_msr(mtrr,  0); /* clear base reg, just for 
                              completeness */ 
      return B_OK; 
    } 
  } 

  /* Generally it is not an error: such range does not exist 
     so it is "disabled" */ 
  return B_ERROR; 
} 

static status_t 
set_memory_range_type(const mtrr_drv_command_t* mtrr_command) 
{ 
  uint32  cr0, cr0_old; 
  uint64  default_mtrr_type;//, mtrr_cap; 
  bool    wc_supported; 
  int     num_mtrrs; 
  status_t  res; 
  int     i; 


  if(!(get_cpu_features() & (1 << 12) ) ) 
    /* MTRRs are not supported */ 
    return B_ERROR; 

  /* Does the CPU support write combining memory ? */ 
//(MSR 0xFE not in VIA CPU)
/*
  mtrr_cap = read_msr(0xFE); 
  wc_supported = ((mtrr_cap & (1 << 10)) != 0 ); 

  if((!wc_supported) &&
     (mtrr_command->mem_type == WC_MEM) &&
     (mtrr_command->action == RANGE_SET)) 
    return B_ERROR; 
*/

  /* How many mtrrs does the CPU have? */ 
//  num_mtrrs = mtrr_cap & 0xFF; 
  num_mtrrs = 8; //(VIA C3 has 8)

  /* Get the default memory type and enable flags for fixed 
     and variable range mtrrs */ 
  default_mtrr_type = read_msr(0x2FF); 

  /* enter no-fill cache mode */ 
  cr0 = cr0_old = get_cr0(); 
  cr0 |=   (1 << 30); 
  cr0 &= ~(1 << 29); 
//  set_cr0(cr0);//(VIA hangs system!!) 

  asm ("wbinvd");  /* flush and invalidate cache */ 
//  set_cr3(get_cr3()); /* flush TLBs */ //(VIA resets system!!)

  write_msr(0x2FF, 0); /* disable all MTRRs */ 
  /* do actual mttr change */ 
  switch(mtrr_command->action) 
  { 
  case RANGE_SET: 
    res = set_mtrr(mtrr_command->base, mtrr_command->len, 
                   mtrr_command->mem_type, 0x200, num_mtrrs); 
    break; 

  case RANGE_CLEAR: 
    clear_mtrr(mtrr_command->base, 0x200, num_mtrrs); 
    res = B_OK; 
    break; 
  } 

  write_msr(0x2FF, default_mtrr_type | (1 << 11) ); 
    /* enable variable-range MTRRs*/ 
  asm ("wbinvd");  /* flush and invalidate cache */ 

 // set_cr3(get_cr3()); /* flush TLBs */ //(VIA resets system!!) 
//  set_cr0(cr0_old);  /* restore caching */ 

  return res; 
} 
